local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-TopItem", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local max, select, pairs = math.max, select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor
local format = string.format
local tbl_remove = table.remove

-- WoW APIs
local CreateFrame, UIParent = CreateFrame, UIParent
local GetItemIcon, GetItemInfo = GetItemIcon, GetItemInfo

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local FRAME_HEIGHT, FRAME_GAP = 40, 2

local MAIN_SOURCE_FLAG_ICONS = GGL.MAIN_SOURCE_FLAG_ICONS
local SUB_SOURCE_FLAG_ICONS = GGL.SUB_SOURCE_FLAG_ICONS
local MAIN_SOURCE_FLAG = GGL.MAIN_SOURCE_FLAG

local ITEM_NUM_FORMAT = "%s x %d"

local GetMoneyText = GGL.GetMoneyTextShort

local function Frame_OnEnter(self)
    if not self.obj.itemData  then return end

    local tooltip = _G.GameTooltip
    tooltip:ClearLines()
    GGL.GUI:SetToolTipPoint(tooltip, self.obj.ttParent, self.obj.ttPoint, 2, 0)
    if self.obj.itemString then
        tooltip:SetHyperlink(self.obj.itemString)
    else
        tooltip:SetItemByID(self.obj.itemID)
    end
    tooltip:Show()
end

local function Frame_OnLeave(self)
    _G.GameTooltip:Hide()
end

--[[-----------------------------------------------------------------------------
PlayerEntry
-------------------------------------------------------------------------------]]
local GetPlayerEntry, ReleasePlayerEntry
do
    -- here are free entrys saved for reuse
    local PlayerEntryRegister = {}

    local PLAYER_WITH_COUNT_F = "(%d) %s"
    local LOOT_ROW_F = "%d x %s"
    local LOOT_ROW_WITH_ICON_F = "|T%s:0|t %d x %s"

    local function BuildTooltipString(lootData)
        -- lootData...
        --ammount, zoneID, sourceID, mainSourceFlag, subSourceFlag
        if not lootData.lootString then
            -- get zoneName
            local mapID, zoneName = GGL:GetZoneData(lootData[2])
            zoneName = zoneName or ""

            -- try get a icon
            local icon
            if lootData[3] and lootData[4] == MAIN_SOURCE_FLAG.ITEM then
                -- from a lockbox
                icon = GetItemIcon(lootData[3])
            end
            if not icon then
                if lootData[5] and SUB_SOURCE_FLAG_ICONS[lootData[5]] then
                    icon = SUB_SOURCE_FLAG_ICONS[lootData[5]]
                elseif lootData[4] and MAIN_SOURCE_FLAG_ICONS[lootData[4]] then
                    icon = MAIN_SOURCE_FLAG_ICONS[lootData[4]]
                end
            end

            -- create the lootString
            if icon then
                lootData.lootString = format(LOOT_ROW_WITH_ICON_F, icon, lootData[1] or 1, zoneName)
            else
                lootData.lootString = format(LOOT_ROW_F, lootData[1] or 1, zoneName)
            end
        end
        return lootData.lootString
    end

    local function PlayerEntry_OnEnter(self)
        if not self.obj.itemLine or not self.obj.playerData then return end

        local playerData = self.obj.playerData
        local tooltip = _G.GameTooltip
        tooltip:ClearLines()
        GGL.GUI:SetToolTipPoint(tooltip, self.obj.itemLine.ttParent, self.obj.itemLine.ttPoint, 2, 0)
        tooltip:AddLine(playerData.playerTemplate.displayName)
        for lootIndex, loot in ipairs(self.obj.playerData.dropData) do
            tooltip:AddLine(BuildTooltipString(loot), 1, 1, 1)
        end
        tooltip:Show()
    end

    local function PlayerEntry_OnLeave(self)
        _G.GameTooltip:Hide()
    end

    local EntryProto = {
        ["OnAcquire"] = function(self)
            self.frame:Show()
        end,

        ["OnRelease"] = function(self)
            self.frame:Hide()
            self:SetPlayerTemplate()
            self:SetItemRowPos()
            self:SetPositionBackground()
        end,

        ["Release"] = function(self)
            ReleasePlayerEntry(self)
        end,

        ["SetPlayerTemplate"] = function(self, playerTemplate)
            if self.playerTemplate == playerTemplate then return end
            self:UnregisterPlayerTemplate()
            if playerTemplate then
                self.playerTemplate = playerTemplate
                playerTemplate:RegisterCallbackEvent("DisplayNameUpdate", self.__onDisplayNameUpdate)
            end
            self:Refresh()
        end,

        ["UnregisterPlayerTemplate"] = function(self)
            if self.playerTemplate then
                self.playerTemplate:UnregisterCallbackEvent("DisplayNameUpdate", self.__onDisplayNameUpdate)
                self.playerTemplate = nil
            end
        end,

        ["SetPlayerData"] = function(self, playerData)
            self.playerData = playerData
        end,

        ["Refresh"] = function(self)
            if self.playerTemplate and self.playerData then
                self.text:SetFormattedText(PLAYER_WITH_COUNT_F, self.playerData.ammount or 1, self.playerTemplate.displayName)
            elseif self.playerTemplate then
                self.text:SetText(self.playerTemplate.displayName)
            else
                self.text:SetText("")
            end
        end,

        ["SetItemRowPos"] = function(self, itemRowPos)
            -- set background color
            if itemRowPos and itemRowPos %2 == 0 then
                self.frameBG:SetColorTexture(0, 0, 0, 0.2)
            else
                self.frameBG:SetColorTexture(1, 1, 1, 0.2)
            end
        end,

        ["SetPositionBackground"] = function(self, playPos)
            if playPos and ( playPos == 2 or playPos+1 %2 == 0 ) then
                self.frameBG:Show()
            else
                self.frameBG:Hide()
            end
        end,
    }
    local EntryProto_mt = {__index = EntryProto}

    function GetPlayerEntry(itemLine)
        local entry = next(PlayerEntryRegister)

        if not entry then
            local frame = CreateFrame("Frame", nil, UIParent)
            frame:SetScript("OnEnter", PlayerEntry_OnEnter)
            frame:SetScript("OnLeave", PlayerEntry_OnLeave)
            frame:SetHeight(12)

            -- background image for list position
            local frameBG = frame:CreateTexture(nil, "BACKGROUND", nil, -1)
            frameBG:SetAllPoints(frame)
            frameBG:SetColorTexture(1, 1, 1, 0.2)
            frameBG:Hide()
            frameBG = frameBG

            local text = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
            text:SetPoint("LEFT", frame, "LEFT", 0, 0)
            text:SetPoint("RIGHT", frame, "RIGHT", -4, 0)
            text:SetHeight(12)
            text:SetJustifyH("LEFT")
            text:SetJustifyV("MIDDLE")

            local function __onDisplayNameUpdate(self)
                entry:Refresh()
            end

            entry = {
                frame = frame,
                frameBG = frameBG,
                text = text,
                __onDisplayNameUpdate = __onDisplayNameUpdate,
            }
            frame.obj = entry

            setmetatable(entry, EntryProto_mt)
        else
            PlayerEntryRegister[entry] = nil
        end

        entry.itemLine = itemLine

        entry.frame:ClearAllPoints()
        entry.frame:SetParent(itemLine.playerContainerFrame)

        entry:OnAcquire()

        return entry
    end

    function ReleasePlayerEntry(playerEntry)
        playerEntry:OnRelease()
        PlayerEntryRegister[playerEntry] = true
    end
end

local function SetPlayerPosition(self, playerRow, pos)
    if pos == 1 then
        -- first one
        playerRow.frame:SetPoint("TOPLEFT", self.playerContainerFrame, "TOPLEFT", 0, 0)
        playerRow.frame:SetPoint("RIGHT", self.playerContainerFrame, "CENTER", -0.5, 0)
    elseif pos % 2 == 0 then
        -- right
        playerRow.frame:SetPoint("TOPLEFT", self.PlayerRows[pos-1].frame, "TOPRIGHT", 0.5, 0)
        playerRow.frame:SetPoint("RIGHT", self.playerContainerFrame, "RIGHT", 0, 0)
    else
        -- next line
        playerRow.frame:SetPoint("TOPLEFT", self.PlayerRows[pos-2].frame, "BOTTOMLEFT", 0, -1)
        playerRow.frame:SetPoint("RIGHT", self.playerContainerFrame, "CENTER", -0.5, 0)
    end

    -- adjust height
    local newSize = 33 + ((pos/2)*12)
    newSize = newSize < FRAME_HEIGHT and FRAME_HEIGHT or newSize
    self.frame:SetHeight(newSize)
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
    ["OnAcquire"] = function(self)
        self:ResetToolTipPoint()
        self:SetListPosition()
        self:SetItemData()
        self.frame:SetHeight(FRAME_HEIGHT)
        self.frame:Show()
    end,

    ["OnRelease"] = function(self)
        self:ClearPlayerList()
        self.frame:Hide()
	end,

    ["OnWidthSet"] = function(self, width)

    end,

    ["SetToolTipPoint"] = function(self, ttPoint, parent)
        self.ttPoint = ttPoint or "BOTTOM_X"
        self.ttParent = parent or self.frame
    end,

    ["ResetToolTipPoint"] = function(self)
        self.ttPoint = "BOTTOM_X"
        self.ttParent = self.frame
    end,

    --- Updates the item with the item data
    ["UpdateItem"] = function(self)
        if not self.itemData then
            self.itemID = nil
            self.itemString = nil
            self:ClearPlayerList()
            return
        end

        local itemData = self.itemData

        local itemName, itemLink = GetItemInfo(itemData.itemString or itemData.itemID)
        self.itemNameText:SetFormattedText(ITEM_NUM_FORMAT, itemLink or "", tonumber(itemData.ammount) or 0)

        -- some things only change with new itemID
        if not self.itemID or self.itemID ~= itemData.itemID then
            self.itemIcon:SetTexture(GetItemIcon(itemData.itemID))
            self.itemMoneyText:SetFormattedText(GetMoneyText(itemData.value))
            self.playerContainerFrame:SetPoint("RIGHT", self.itemMoneyText, "RIGHT", -(self.itemMoneyText:GetStringWidth() + 5), 0)

            self.itemID = itemData.itemID
            self.itemString = itemData.itemString
        end
    end,

    --- Adds a player and/or updates
    ["AddPlayer"] = function(self, playerData)
        local playerRow
        -- search player
        for i = 1, #self.PlayerRows do
            if self.PlayerRows[i].playerData.playerKey == playerData.playerKey then
                playerRow = self.PlayerRows[i]
                break
            end
        end

        if not playerRow then
            playerRow = GetPlayerEntry(self)
            self.PlayerRows[#self.PlayerRows+1] = playerRow

            local rowPos = #self.PlayerRows

            SetPlayerPosition(self, playerRow, rowPos)

            playerRow:SetPlayerData(playerData)
            -- Must be fixed an what colors?
            --playerRow:SetItemRowPos(self.listPosition)
            --playerRow:SetPositionBackground(rowPos)
        end

        playerRow:SetPlayerTemplate(playerData.playerTemplate)

        playerRow:Refresh()
    end,

    -- resort players
    ["ResortPlayers"] = function(self)
        for i = 1, #self.PlayerRows do
            SetPlayerPosition(self, self.PlayerRows[i], i)
        end
    end,

    --- Adds a player and/or updates
    ["RemovePlayer"] = function(self, playerData)
        -- search player
        for i = 1, #self.PlayerRows do
            if self.PlayerRows[i].playerData.playerKey == playerData.playerKey then
                self.PlayerRows[i]:Release()
                tbl_remove(self.PlayerRows, i)
                self:ResortPlayers()
                return
            end
        end
    end,

    --- Sets the item data from TopItems
    ["SetItemData"] = function(self, itemData)
        self.itemData = itemData
        self:UpdateItem()
    end,

    --- Shows a background when pos % 2
    ["SetListPosition"] = function(self, pos)
        if pos and pos % 2 == 0 then
            self.frameBG:Show()
        else
            self.frameBG:Hide()
        end
        self.listPosition = pos
    end,

    --- Register events for a player
    ["RegisterPlayerEvents"] = function(self, playerKey, playerTemplate)
        if not self.PlayerRows then return end

        local playerRow
        for i = 1, #self.PlayerRows do
            if self.PlayerRows[i].playerData.playerKey == playerKey then
                self.PlayerRows[i]:SetPlayerTemplate(playerTemplate)
                return
            end
        end
    end,

    --- Unregister events for a player
    ["UnregisterPlayerEvents"] = function(self, playerTemplate)
        if not self.PlayerRows then return end

        local playerRow
        for i = 1, #self.PlayerRows do
            if self.PlayerRows[i].playerData.playerTemplate == playerTemplate then
                self.PlayerRows[i]:UnregisterPlayerTemplate()
                return
            end
        end
    end,

    --- Unregister all player events
    ["UnregisterAllPlayerEvents"] = function(self)
        if not self.PlayerRows then return end
        for i = 1, #self.PlayerRows do
            self.PlayerRows[i]:UnregisterPlayerTemplate()
        end
    end,

    --- Unregister all player events
    ["ClearPlayerList"] = function(self)
        if not self.PlayerRows then return end
        for i = 1, #self.PlayerRows do
            self.PlayerRows[i]:Release()
            self.PlayerRows[i] = nil
        end
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("Frame", nil, UIParent)
    frame:SetHeight(FRAME_HEIGHT)
    frame:SetScript("OnEnter", Frame_OnEnter)
    frame:SetScript("OnLeave", Frame_OnLeave)
    frame.obj = widget
    widget.frame = frame

    -- background image for list position
    local frameBG = frame:CreateTexture(nil, "BACKGROUND", nil, -1)
    frameBG:SetAllPoints(frame)
    frameBG:SetColorTexture(1, 1, 1, GGL.LINE_ALPHA_VALUE)
    frameBG:Hide()
    widget.frameBG = frameBG

    -- itemIcon
    local itemIcon = frame:CreateTexture(nil, "BACKGROUND")
    itemIcon:SetPoint("TOPLEFT", frame, "TOPLEFT", FRAME_GAP, -FRAME_GAP)
    itemIcon:SetSize(FRAME_HEIGHT-(FRAME_GAP*2), FRAME_HEIGHT-(FRAME_GAP*2))
    widget.itemIcon = itemIcon

    -- itemNameText
    local itemNameText = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
    itemNameText:SetPoint("TOP", frame, "TOP", 0, -FRAME_GAP)
    itemNameText:SetPoint("LEFT", itemIcon, "RIGHT", FRAME_GAP, 0)
    itemNameText:SetPoint("RIGHT", frame, "RIGHT", -FRAME_GAP, 0)
    itemNameText:SetHeight(15)
    itemNameText:SetJustifyH("LEFT")
    itemNameText:SetJustifyV("MIDDLE")
    widget.itemNameText = itemNameText

    -- itemMoneyText
    local itemMoneyText = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
    itemMoneyText:SetPoint("TOPLEFT", itemNameText, "BOTTOMLEFT", 0, -FRAME_GAP)
    itemMoneyText:SetPoint("TOPRIGHT", itemNameText, "BOTTOMRIGHT", 0, -FRAME_GAP)
    itemMoneyText:SetPoint("BOTTOM", itemIcon, "BOTTOM", 0, 0)
    itemMoneyText:SetJustifyH("RIGHT")
    itemMoneyText:SetJustifyV("TOP")
    widget.itemMoneyText = itemMoneyText

    local playerContainerFrame = CreateFrame("Frame", nil, frame)
    playerContainerFrame:SetPoint("TOPLEFT", itemNameText, "BOTTOMLEFT", 0, -FRAME_GAP)
    playerContainerFrame:SetPoint("RIGHT", itemMoneyText, "RIGHT", 0, 0)
    playerContainerFrame:SetPoint("BOTTOM", frame, "BOTTOM", 0, FRAME_GAP)
    playerContainerFrame:SetScript("OnEnter", function() end)
    widget.playerContainerFrame = playerContainerFrame

    widget.PlayerRows = {}

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)