local AddonName, GGL = ...
local Templates = GGL.Templates

local Session, SessionHelper = Templates:Register("Session")

-- lua
local tonumber = tonumber
local pairs, ipairs = pairs, ipairs
local format, match = string.format, string.match
local time = time

-- wow
local C_TimerAfter = C_Timer.After
local GetBestMapForUnit = C_Map.GetBestMapForUnit

-- ##################
-- local
-- ##################
local RunningSessions = {}

local SESSION_FORMAT = "S%d"
local TIMER_INTERVAL = 1

-- lootTime, zoneIniID, difficultyID, itemID, ammount, sourceID, mainSourceFlag, subSourceFlag
local ITEM_DB_FORMAT = "%d\001%d\001%d\001%s\001%d\001%d\001%d\001%d"
local ITEM_DB_FORMAT_MATCH = "(%d*)\001(%d*)\001(%d*)\001(.*)\001(%d*)\001(%d*)\001(%d*)\001(%d*)"
local PLAYER_FORMAT = "%s:%s"

local function AddLogString(self, playerTemplate, event, source, db, lootTime, zoneID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, pauseLog)
    if pauseLog or self.pauseLog or lootTime < 0 then return end
    itemLog = tostring(itemLog)

    if source then
        -- source(mainSourceFlag, sourceID, ammount)
        for index, src in ipairs(source) do
            if event == "AddLogMoney" and #source <= 1 and src[3] <= 0 then
                db[ #db + 1 ] = format(ITEM_DB_FORMAT, lootTime, zoneID, difficultyID, itemLog, ammount, (sourceID or src[2]) or 0, src[1] or mainSourceFlag, subSourceFlag)
                self:FireCallbackEvent(event, self, playerTemplate, db[ #db ], #db)
                break
            else
                db[ #db + 1 ] = format(ITEM_DB_FORMAT, lootTime, zoneID, difficultyID, itemLog, src[3], (sourceID or src[2]) or 0, src[1] or mainSourceFlag, subSourceFlag)
                self:FireCallbackEvent(event, self, playerTemplate, db[ #db ], #db)
            end
        end
    else
        -- MainWindow.frame:AddItem(itemID, sourceID, ammount, MainWindow.ItemPrice:GetPrice(itemID) * ammount, mainSourceFlag, subSourceFlag)
        db[ #db + 1 ] = format(ITEM_DB_FORMAT, lootTime, zoneID, difficultyID, itemLog, ammount, sourceID or 0, mainSourceFlag, subSourceFlag)
        self:FireCallbackEvent(event, self, playerTemplate, db[ #db ], #db)
    end
end

-- ##################
-- static
-- ##################
local Static = {}
Templates.Session = Static

-- combine by mainSourceFlag
function Static:FormatSource(source)
    local tmp = {}
    -- source(mainSourceFlag, sourceID, ammount)
    for index, src in ipairs(source) do
        tmp[src[1]] = ( tmp[src[1]] or 0 ) + src[3]
    end
    return tmp
end

function Static:GetRunningSession(sessionDB)
    sessionDB = sessionDB or GGL.sessionDb.Session
    if sessionDB.RunningSession then
        return Templates:New("Session", sessionDB, sessionDB.RunningSession)
    end
end

function Static:CloseOpenSessions()
    for k,v in pairs(RunningSessions) do
        v:Close()
    end
end

function Static:GetDataFromLogString(logString)
    local lootTime, zoneIniID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag = match(logString, ITEM_DB_FORMAT_MATCH)
    local itemID = GGL:GetItemDataFromItemLog(itemLog)
    return tonumber(lootTime), tonumber(zoneIniID), tonumber(difficultyID), tonumber(itemID), itemLog, tonumber(ammount), tonumber(sourceID), tonumber(mainSourceFlag), tonumber(subSourceFlag)
end

-- ##################
-- template
--- Values
-- isClosed             Session closed and unloaded
-- isRunning            Session is running
-- durationTime         Duration of the session
-- startTime            Start time of the session
-- endTime              End time of the session
-- set                  Set key string
-- dataIsSet            Data is set and session can be started

-- EVENTS
--      "AddItem"       (sessionSelf, itemID, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
--      "AddMoney"      (sessionSelf, copper, sourcePlayer)
--      "AddLogItem"    (sessionSelf, playerTemplate, logString, logCounter)
--      "AddLogMoney"   (sessionSelf, playerTemplate, logString, logCounter)
--      "OnClose"       (sessionSelf)
--      "OnStop"        (sessionSelf)
--      "SetDataRefresh"(sessionSelf)
-- ##################

Session.sessionFormat = "S%d"
Session.sessionType = "S"

--- New session or reload one
function Session:Constructor(sessionDB, sessionID)
    if not sessionDB or type(sessionDB) ~= "table" then return end

    -- check for running session or create new one
    if not sessionID then
        sessionID = format(self.sessionFormat, GGL.GetTime())
    end
    if RunningSessions[sessionID] then
        return RunningSessions[sessionID]
    end
    if not sessionDB[sessionID] then
        sessionDB[sessionID] = {}
    end

    -- set template variables
    self.dbRaw = sessionDB
    self.db = sessionDB[sessionID]
    -- set db variables
    self.db.sessionID = sessionID   -- double the fun
    self.dbRaw.RunningSession = sessionID
    -- add own callback handler
    GGL:AddCallbackHandler(self)

    local db = self.db

    -- if its a running session resume or end it
    if (db.isRunning or db.isStopped) or (db.endTime and db.endTime <= GGL.GetTime()) then
        if db.endTime and db.endTime <= GGL.GetTime() then
            return self:Close()
        else
            self:SetData()
        end
    elseif db.isClosed then
        return self:Close()
    else
        -- new one
    end

    -- list for syncID's
    self.mergeIDList = {}

    -- create lootLog table
    db.lootLog = db.lootLog or {}

    -- set own bnet
    self.selfPlayer = GGL.Templates.BNWoWPlayer:GetSelfPlayer()
    self.selfDbKey = self:GetPlayerSyncID(self.selfPlayer)
    if not db.lootLog[self.selfDbKey] then db.lootLog[self.selfDbKey] = {} end
    self.lootLog = db.lootLog[self.selfDbKey]

    -- Register loot module
    self.OnItemLoot_CB = function(Loot, itemID, itemLog, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
        if not db.isRunning then return end
        self:AddNewItem(itemID, itemLog, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
    end

    self.OnMoneyLoot_CB = function(Loot, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
        if not db.isRunning then return end
        self:AddMoney(ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
    end

    -- Register zone change
    self.OnMapChange_CB = function(sender, mapID, mapName, mapTypeFlag)
        self.zoneID = mapID
    end
    self.OnMapChange_CB(nil, GGL:GetZoneData())

    -- Set Price table
    if not db.priceLog then db.priceLog = {} end

    -- Add into statistics
    self.statistics = GGL.Statistics:RegisterSession(self)

    -- db instanceResetData
    self.db.instanceResetData = self.db.instanceResetData or {}

    -- register as running
    RunningSessions[sessionID] = self
end

function Session:GetPlayerSyncID()
    -- the player gets a fixed mergeID
    -- TODO: #9 Add more merge settings here
    return self.selfPlayer.syncID
end

--- Returns the merge ID for a playerTemplate
--- For session that returns just "self"
function Session:GetPlayerMergeID(playerTemplate)
    return "self"
end

--- Returns the active player for a mergeID
--- For session that returns just the player
-- @return  playerTemplate
function Session:GetActivePlayerForMergeID(mergeID)
    return self.selfPlayer
end

--- Returns the merge list
-- format: mergeList = { [playerSyncID] = <0=noID or "string"=mergeID> }
function Session:GetMergeList()
    return self.mergeIDList
end

function Session:GetDataFromLogString(logString)
    local lootTime, zoneIniID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag = match(logString, ITEM_DB_FORMAT_MATCH)
    local itemID = GGL:GetItemDataFromItemLog(itemLog)
    return tonumber(lootTime), tonumber(zoneIniID), tonumber(difficultyID), tonumber(itemID), itemLog, tonumber(ammount), tonumber(sourceID), tonumber(mainSourceFlag), tonumber(subSourceFlag)
end

function Session:AddMoney(ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag, zoneID, difficultyID, lootTime, lootLog, sourcePlayer, noLog)
    if not self.dataIsSet then return end
    local db = self.db
    -- lootLog(dbTable) and lootTime check
    lootLog = lootLog or self.lootLog
    lootTime = lootTime or GGL.GetTime()
    lootTime = lootTime - db.startTime
    -- get loot zone
    zoneID = zoneID or self.zoneID
    difficultyID = difficultyID or self.zoneDifficultyID

    sourcePlayer = sourcePlayer or self.selfPlayer

    -- Add loot string
    AddLogString(self, self.selfPlayer, "AddLogMoney", source, lootLog, lootTime, zoneID, difficultyID, 0, ammount, sourceID, mainSourceFlag, subSourceFlag, noLog)

    -- check is is valid
    if (self.ItemPrice and self.ItemPrice:RawGoldAllowed()) and (not self.ZoneFilter or self.ZoneFilter:IsValidZone(zoneID, difficultyID)) then
        if source then
            source = Static:FormatSource(source)
            for k, v in pairs(source) do
                if not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(k or mainSourceFlag, subSourceFlag) then
                    self:FireCallbackEvent("AddMoney", self, ammount, sourcePlayer)
                    break
                end
            end
        else
            if not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(mainSourceFlag, subSourceFlag) then
                self:FireCallbackEvent("AddMoney", self, ammount, sourcePlayer)
            end
        end
    end
end

function Session:AddNewItem(itemID, itemLog, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag, zoneID, difficultyID, lootTime, lootLog, sourcePlayer, noLog)
    if not self.dataIsSet then return end
    local db = self.db
    -- lootLog and lootTime check
    lootLog = lootLog or self.lootLog
    lootTime = lootTime or time()
    lootTime = lootTime - db.startTime
    -- save price for the log
    local newItemID, itemString = GGL:GetItemDataFromItemLog(itemLog)
    local price = self.ItemPrice:GetPrice(newItemID, itemString)
    price = price or 0

    -- get loot zone
    zoneID = zoneID or self.zoneID
    difficultyID = difficultyID or self.zoneDifficultyID

    sourcePlayer = sourcePlayer or self.selfPlayer

    -- always into log
    AddLogString(self, self.selfPlayer, "AddLogItem", source, lootLog, lootTime, zoneID, difficultyID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag, noLog)

    -- check is is valid
    local itemFilterPass = false
    if (self.ItemFilter and self.ItemFilter:IsValidItem(itemID)) and (not self.ZoneFilter or self.ZoneFilter:IsValidZone(zoneID, difficultyID)) then
        if source then
            source = Static:FormatSource(source)
            for k, v in pairs(source) do
                if not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(k or mainSourceFlag, subSourceFlag) then
                    self:FireCallbackEvent("AddItem", self, newItemID, itemString, zoneID, sourceID, v, price, k or mainSourceFlag, subSourceFlag, sourcePlayer)
                    GGL.LootAlert:ProcessAlerts(newItemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, price, sourcePlayer)
                    itemFilterPass = true
                end
            end
        else
            if not self.LootSourceFilter or self.LootSourceFilter:IsValidSource(mainSourceFlag, subSourceFlag) then
                self:FireCallbackEvent("AddItem", self, newItemID, itemString, zoneID, sourceID, ammount, price, mainSourceFlag, subSourceFlag, sourcePlayer)
                GGL.LootAlert:ProcessAlerts(newItemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, price, sourcePlayer)
                itemFilterPass = true
            end
        end
    end
    return itemFilterPass
end

function Session:AddFromLogString(logString)
    self.pauseLog = true
    local lootTime, zoneIniID, difficultyID, itemID, itemLog, ammount, sourceID, mainSourceFlag, subSourceFlag = self:GetDataFromLogString(logString)
    if itemID == 0 then
        self:AddMoney(ammount, nil, sourceID, mainSourceFlag, subSourceFlag, nil, zoneIniID, difficultyID, lootTime)
    else
        GGL:IsItemCached(itemID, self.AddNewItem, self, itemID, itemLog, ammount, nil, sourceID, mainSourceFlag, subSourceFlag, nil, zoneIniID, difficultyID, lootTime, nil, nil, true)
    end
    self.pauseLog = nil
end

function Session:RefreshSettingsSet(setRefresh, priceRefresh)
    local db = self.db
    if db.setSave then
        if setRefresh and self.ItemPrice then
            self.ItemPrice:LoadSetSettings(db.setSave.itemPrice, self.db.priceLog)
            self.ItemFilter:LoadSetSettings(db.setSave.ItemFilter)
            self.ZoneFilter:LoadSetSettings(db.setSave.zoneFilter, db.setSave.instanceFilter)
            self.LootSourceFilter:LoadSetSettings(db.setSave.lootSourceFilter)
        else
            self.ItemPrice = Templates:New("ItemPriceSet", db.setSave.itemPrice, true, self.db.priceLog)
            self.ItemFilter = Templates:New("ItemFilter", db.setSave.ItemFilter, true)
            self.ZoneFilter = Templates:New("ZoneFilter", db.setSave.zoneFilter, db.setSave.instanceFilter, true)
            self.LootSourceFilter = Templates:New("LootSourceFilter", db.setSave.lootSourceFilter, true)
        end
    end
    -- refresh cached prices
    wipe(db.priceLog)
    if priceRefresh then
        for itemID in pairs(db.priceLog) do
            self.ItemPrice:GetPrice(GGL:GetItemDataFromItemLog(itemID))
        end
    end
    -- check if we have the mainframe
    if self.mainFrame then
        self:AddIntoMainFrame(self.mainFrame)
    end
end

function Session:SetData(set, setRefresh)
    -- change data of running session not allowed
    local db = self.db

    -- Setdata
    set = set or "Base"
    if type(set) == "table" then
        db.setSave = set
        db.set = false
    else
        if set == true or db.set == false then

        elseif not db.set or setRefresh then
            local SetData = Templates:New("Set", set)
            db.setSave = {
                itemPrice = SetData:GetPriceSource(true),
                ItemFilter = SetData:GetItemFilter(true),
                zoneFilter = SetData:GetZoneFilter(true),
                instanceFilter = SetData:GetInstanceFilter(true),
                lootSourceFilter = SetData:GetLootSourceFilter(true),
            }
        end
        db.set = db.set or set
    end

    -- set player name
    self.playerName = format("%s-%s", UnitName("player"),  GetNormalizedRealmName())

    -- Register/Refresh set
    self:RefreshSettingsSet(setRefresh)

    self.db.dataIsSet = true
    self.dataIsSet = true
end

function Session:ReloadSet()
    if self.db.setSave then
        self:RefreshSettingsSet(true, true)

        -- Fire event before the mainFrame refresh
        -- so other modules get the AddFromLogString from mainFrame also
        self:FireCallbackEvent("SetDataRefresh", self)

        if self.mainFrame then
            self:RemoveFromMainFrame()
            self:AddIntoMainFrame(true)
            self:RebuildMainFrameList()
        end
    end
end

function Session:MainFrameOnNewItem(itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
    if not self.mainFrame or sourcePlayer ~= self.selfPlayer then return end
    self.mainFrame:AddItem(itemID, self:LoggedItemUseVariations(itemID, itemString) and itemString or nil, sourceID, ammount, copperValue * ammount, mainSourceFlag, subSourceFlag)
    self.mainFrame:SetItemValueGold(copperValue * ammount, true)
end

function Session:MainFrameOnAddMoney(copper, sourcePlayer)
    if not self.mainFrame or sourcePlayer ~= self.selfPlayer then return end
    self.mainFrame:SetRawGold(copper, true)
end

function Session:RebuildMainFrameList()
    if not self.mainFrame or self.rebuild then return end
    GGL.LootAlert:Pause(true)
    for playerSyncID, lootData in pairs(self.db.lootLog) do
        -- ignore empty tables
        if #lootData > 0 then
            local player = Templates.Sync.GetPlayerFromSync(playerSyncID, true)
            -- only player
            if player.isPlayerBattleTag then
                for i = 1, #lootData do
                    self:AddFromLogString(lootData[i])
                end
            end
        end
    end
    self.rebuild = true
    GGL.LootAlert:Pause(false)
end

function Session:AddIntoMainFrame(add)
    if not self.db or self.db.isClosed then return end
    if add then
        local frame = GGL.GUI.MainWindow.frame or GGL.GUI.MainWindow:CreateFrame()
        frame:Clear()
        local db = self.db
        -- start timer if its a running session or just set timer
        if self.db.endlessTime  then
            if db.isRunning  then
                frame:SetEndlessTimer(self.db.timeElapsed, self.inviteTemplate)
                frame.contentContainer.timer:StartTimer()
                frame:Show()
            end
        elseif db.isRunning then
            frame:SetEndlessTimer(nil)
            frame.contentContainer.timer:SetTime(db.startTime, db.endTime)
            frame.contentContainer.timer:IsCounter(false)
            frame.contentContainer.timer:StartTimer()
            frame:Show()
        elseif db.durationTime then
            frame:SetEndlessTimer(nil)
            frame.contentContainer.timer:SetDuration(db.durationTime)
        end

        -- set session
        frame:SetSession(self)

        -- set zone filter and refresh zone
        frame.ZoneFilter = self.ZoneFilter
        frame:RefreshMap()

        -- set data
        self.mainFrame = frame
        frame.runningSession = self

        -- callback
        self:RegisterCallbackEvent("AddItem", Session.MainFrameOnNewItem)
        self:RegisterCallbackEvent("AddMoney", Session.MainFrameOnAddMoney)

        -- add into minimap button
        GGL.GUI.MainWindow:AddIntoMiniMapButton()
    else
        self:RemoveFromMainFrame()
        -- only hide mainFrame when session is running
        -- that way the frame is still open when the session ends but gets hidden if the state changes mid session
        if self.db.isRunning then
            GGL.GUI.MainWindow:CreateFrame():Hide()
        end
    end
end

function Session:RemoveFromMainFrame()
    if not self.mainFrame then return end
    local frame = self.mainFrame
    self.mainFrame = nil

    frame:SetSession(nil)

    frame.ZoneFilter = nil
    if self.db.isClosed then
        frame.contentContainer.timer:FinishTimer()
    else
        frame.contentContainer.timer:FinishTimer(true)
    end

    frame:RefreshMap()
    frame.runningSession = nil

    self:UnregisterCallbackEvent("AddItem", Session.MainFrameOnNewItem)
    self:UnregisterCallbackEvent("AddMoney", Session.MainFrameOnAddMoney)

    self.rebuild = nil

    -- remove from minimap button
    GGL.GUI.MainWindow:RemoveFromMiniMapButton()
end

function Session:SetTimerData(startTime, endTime)
    if self.db.startTime then return end
    self.db.startTime = startTime
    self.db.endTime = endTime
    if endTime then
        self.db.durationTime = endTime - startTime
    end
end

function Session:SetTimerDuration(duration)
    local curTime = GGL.GetTime()
    self:SetTimerData(curTime, curTime + duration)
end

function Session:SetTimerEndless()
    self.db.endlessTime = true
    self.db.timeElapsed = self.db.timeElapsed or 0
    self.db.startTime = GGL.GetTime() - self.db.timeElapsed
end

function Session:UpdateResetData(playeTemplate, data)
    if not data and playeTemplate ~= self.selfPlayer then return end

    local enterData, resetData
    if playeTemplate == self.selfPlayer then
        enterData = GGL:InstanceResets_GetInstanceEnterTimes(self:GetStartTime() - 1800, true)
        resetData = GGL:InstanceResets_GetInstanceResetTimes(self:GetStartTime())
    end

    local playerMergeID = self:GetPlayerMergeID(playeTemplate) or playeTemplate.syncID

    if not self.db.instanceResetData[playerMergeID] then
        self.db.instanceResetData[playerMergeID] = {} -- {mapID,enterTime,resetTime}
    end

    if data then
        self.db.instanceResetData[playerMergeID] = data
        return data
    end

    local playerData = self.db.instanceResetData[playerMergeID]

    local idList = {}
    for i = 1, #playerData do
        idList[playerData[i][1]..playerData[i][2]] = playerData[i]
    end

    local doResort = false
    for i = 1, #enterData do
        local data = enterData[i]
        if not idList[(data.mapID or data.mapName)..data.time] then
            playerData[#playerData+1] = {
                data.mapID or data.mapName,
                data.time,
            }
        end
    end

    for i = 1, #resetData do
        local data = resetData[i]
        if idList[(data.mapID or data.zone)..data.enterTime] then
            idList[(data.mapID or data.zone)..data.enterTime][3] = data.startTime
        else
            playerData[#playerData+1] = {
                data.mapID or data.zone,
                data.enterTime,
                data.startTime
            }
            doResort = true
        end
    end

    if doResort then
        table.sort(playerData, function (a, b)
            return a[2] < b[2]
        end)
    end

    return playerData
end

function Session:UpdateOwnResetData(noCallback)
    local data = self:UpdateResetData(self.selfPlayer)
    if not noCallback then
        self:FireCallbackEvent("OnInstancesResetUpdate", self, self:UpdateResetData(self.selfPlayer))
    end
    return data
end


function Session:Start()
    local db = self.db
    if db.isClosed or not db.dataIsSet then return end

    if not db.startTime then
        error(format("time not set for '%s'", self.sessionID))
    end

    -- statistics
    self.statistics:StartTracking()

    -- register events
    GGL.Loot:RegisterCallbackEvent("OnItemLoot", self.OnItemLoot_CB)
    GGL.Loot:RegisterCallbackEvent("OnMoneyLoot", self.OnMoneyLoot_CB)
    GGL:RegisterCallbackEvent("OnMapChange", self.OnMapChange_CB)

    db.isRunning = true
    db.isStopped = nil
    self.isRunning = true

    -- create the end timer
    if not self.timerUpdate and not self.db.endlessTime then
        self.timerUpdate = function()
            if db.endTime <= GGL.GetTime() then
                self:Close()
            else
                C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
            end
        end
        C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
    elseif self.db.endlessTime then
        self:SetTimerEndless()
        self.timerUpdate = function()
            if db.isClosed or not db.isRunning then
                return
            else
                db.timeElapsed = GGL.GetTime() - db.startTime
                C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
            end
        end
        C_TimerAfter(TIMER_INTERVAL, self.timerUpdate)
    end

    if self.mainFrame then
        self:AddIntoMainFrame(true)
        GGL.GUI.MainWindow:Show()
    end

    -- Instance resets
    if not self.cbOnInstancesResetUpdate then
        self.cbOnInstancesResetUpdate = function()
            self:UpdateOwnResetData()
        end
        GGL:RegisterCallbackEvent("OnInstancesResetNewLockout", self.cbOnInstancesResetUpdate)
        GGL:RegisterCallbackEvent("OnInstancesResetEnterInstance", self.cbOnInstancesResetUpdate)
    end

    -- rebuild list from db
    self:RebuildMainFrameList()

    -- local
    GGL.Session:LockSession(self)

    -- update zone for start
    self.OnMapChange_CB(nil, GGL:GetZoneData())

    return db.startTime, db.endTime, db.durationTime
end

local function UnregisterAll(self)
    -- remove from loot callback
    GGL.Loot:UnregisterCallbackEvent("OnItemLoot", self.OnItemLoot_CB)
    GGL.Loot:UnregisterCallbackEvent("OnMoneyLoot", self.OnMoneyLoot_CB)
    GGL:UnregisterCallbackEvent("OnMapChange", self.OnMapChange_CB)

    -- stop statistics
    if self.statistics then
        self.statistics:StopTracking()
    end

    -- stop timer
    self:RemoveFromMainFrame()

    -- unregister all events
    self:UnregisterAllCallbackEvents()

    -- reset events
    if self.cbOnInstancesResetUpdate then
        GGL:UnregisterCallbackEvent("OnInstancesResetNewLockout", self.cbOnInstancesResetUpdate)
        GGL:UnregisterCallbackEvent("OnInstancesResetEnterInstance", self.cbOnInstancesResetUpdate)
        self.cbOnInstancesResetUpdate = nil
    end

    -- remove from running sessions
    RunningSessions[self.db.sessionID] = nil
end

function Session:Unregister()
    UnregisterAll(self)
end

function Session:Stop()
    -- only stop runnig sessions
    if not self.db.isRunning then return end
    local db = self.db
    db.isStopped = true
    db.isRunning = false
    self.isRunning = false

    self:FireCallbackEvent("OnStop", self)

    UnregisterAll(self)

    -- unlock
    GGL.Session:UnlockSession(self)

    RunningSessions[self.db.sessionID] = self
end

--- Close the session
function Session:Close()
    local db = self.db
    db.isClosed = true
    db.isRunning = false
    db.isStopped = nil
    self.isRunning = false

    db.closeTime = GGL.GetTime()

    self:FireCallbackEvent("OnClose", self)

    -- move and delete link
    GGL.sessionDb.Archiv[db.sessionID] = self.dbRaw[db.sessionID]
    self.dbRaw[db.sessionID] = nil

    UnregisterAll(self)

    -- unlock
    GGL.Session:UnlockSession(self)
end

-- ##################
-- Helper
-- ##################
function Session:GetType()
    return SessionHelper:GetType()
end

function Session:IsRunning()
    return self.db.isRunning
end

function Session:IsStopped()
    return self.db.isStopped and true or false
end

function Session:IsRunningTemplate()
    return self.isRunning
end

function Session:IsClosed()
    return self.db.isClosed
end

function Session:GetSet()
    return self.db.set
end

function Session:GetSessionType()
    return self.sessionType
end

function Session:DataIsSet()
    return self.db.dataIsSet
end

function Session:GetTimer()
    return self.db.startTime, self.db.endTime, self.db.durationTime
end

function Session:GetStartTime()
    return self.db.startTime
end

--- Returns session end time
-- For endless sessions that returns 0
function Session:GetEndTime()
    return self.db.endlessTime and 0 or self.db.endTime
end

--- Returns session duration
-- For endless sessions that returns the startTime
function Session:GetDuration()
    return self.db.endlessTime and self.db.startTime or self.db.durationTime
end

function Session:GetSessionID()
    return self.db.sessionID
end

function Session:LoggedItemUseVariations(itemID, itemString)
    if self.ItemPrice then
        return self.ItemPrice:LoggedItemUseVariations(itemID, itemString)
    end
end